@@ -1,44 +1,54 @@
 /***************************************
-  $Header: /home/amb/xbomb/RCS/xwindow.c 1.14 1995/11/15 19:26:42 amb Exp $
+  $Header: /home/amb/xbomb/RCS/xwindow.c 1.23 1999/04/10 18:28:36 amb Exp $
 
-  XBomb - 'Minesweeper' game - Version 2
+  XBomb - 'Minesweeper' game - Version 2.1a.
 
   All X Window functions.
-
+  ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1994 1995 Andrew M. Bishop
+  This file Copyright 1994,1995,1998 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
   ***************************************/
 
 
+#include <stdio.h>
+
 #include <X11/X.h>
 #include <X11/Xlib.h>
 #include <X11/Intrinsic.h>
 #include <X11/StringDefs.h>
 #include <X11/Xaw/Command.h>
+#include <X11/Xaw/Toggle.h>
 #include <X11/Xaw/Form.h>
 #include <X11/Xaw/Label.h>
 #include <X11/Xaw/MenuButton.h>
 #include <X11/Xaw/SimpleMenu.h>
 #include <X11/Xaw/SmeBSB.h>
 #include <X11/keysym.h>
-#include <stdio.h>
+
 #include <sys/time.h>
 #include <unistd.h>
 
 #include "xbomb.h"
 #include "icon.h"
 
-#if defined(__sun__)
+#if defined(__sun__) && !defined(__svr4__)
 int gettimeofday(struct timeval *tv, struct timezone *tz);
 #endif
 
-extern int grid_width,grid_height,grid_bombs,grid_type;
+/*+ The current information about the grid +*/
+extern int grid_width,          /*+ width. +*/
+           grid_height,         /*+ height. +*/
+           grid_bombs,          /*+ number of bombs. +*/
+           grid_type,           /*+ type. +*/
+           grid_level;          /*+ level. +*/
 
-extern char levels[NLEVELS][LEVELSLEN],types[NTYPES][TYPESLEN];
+/*+ The names of the different game +*/
+extern char *levels[NLEVELS],   /*+ levels (difficulty). +*/
+            *types[NTYPES];     /*+ types (grid shapes). +*/
 
 static int xoffset,yoffset;
 static int status=0;
@@ -47,7 +57,8 @@ static int unsigned pixw,pixh;
 static XtAppContext app_context;
 static Display* display;
 static Window play_window;
-static Widget play_area,clock_w,uxb_w;
+static Widget play_area,clock_w,uxb_w,hiscore,toplevel;
+static Widget highscore,highscoresform,highscores[12][4];
 static Pixmap icon_p,bombpix[2]={0,0};
 static XPoint outline[8];
 
@@ -56,40 +67,42 @@ static void mouse_press_proc(Widget w,XtPointer va,XEvent* e,Boolean* vb);
 static void key_press_proc(Widget w,XtPointer va,XEvent* e,Boolean* vb);
 static void change_status_proc(Widget widget,XtPointer clientData,XtPointer callData);
 static void hiscore_proc(Widget widget,XtPointer clientData,XtPointer callData);
+static void close_proc(Widget w,XtPointer va,XEvent* e,Boolean* vb);
 static void display_clock(XtPointer p,XtIntervalId i);
 static int msecs(void);
 
-/*+ An array to hold the colour and font resource data in. +*/
+/*+ A structure to hold the colour and font resource data in. +*/
 struct temp
 {
- Pixel colours[6];
- XFontStruct* fontstruct;
+ Pixel colours[6];              /*+ The colours. +*/
+ XFontStruct* fontstruct;       /*+ The font. +*/
 }
 resources;
 
 /*+ An array to hold the GCs in. +*/
 static GC gc[7];
 
-#define GC_BOMB        0
-#define GC_ACTUAL_BACK 1
-#define GC_THINK_BACK  2
-#define GC_NUMBERS     3
-#define GC_UNSEEN      4
-#define GC_CORRECT     5
-#define GC_BLANK       6
+#define GC_BOMB        0        /*+ For drawing the bomb and flag foreground. +*/
+#define GC_ACTUAL_BACK 1        /*+ For drawing the bomb background. +*/
+#define GC_THINK_BACK  2        /*+ For drawing the flag background. +*/
+#define GC_NUMBERS     3        /*+ For drawing the numbers on the grid. +*/
+#define GC_UNSEEN      4        /*+ For unseen (hidden) tiles. +*/
+#define GC_CORRECT     5        /*+ For correctly guessed tiles at the end.+*/
+#define GC_BLANK       6        /*+ For blank squares. +*/
 
 /*+ The colour resources for the map. +*/
 static XtResource xresources[]=
 {
- {"bomb"     , "XBomb"  , XtRPixel  , sizeof(String)  , 0                , XtRString, "black" },
- {"bombreal" , "XBomb"  , XtRPixel  , sizeof(String)  ,   sizeof(String) , XtRString, "red" },
- {"bombmark" , "XBomb"  , XtRPixel  , sizeof(String)  , 2*sizeof(String) , XtRString, "orange" },
- {"number"   , "XBomb"  , XtRPixel  , sizeof(String)  , 3*sizeof(String) , XtRString, "blue" },
- {"hidden"   , "XBomb"  , XtRPixel  , sizeof(String)  , 4*sizeof(String) , XtRString, "grey50" },
- {"correct"  , "XBomb"  , XtRPixel  , sizeof(String)  , 5*sizeof(String) , XtRString, "green" },
- {"font"     , "XBomb"  , XtRFontStruct , sizeof(XFontStruct*) , 6*sizeof(String), XtRString, "lucidasans-24"}
+ {"bomb"    , "XBomb", XtRPixel     , sizeof(String)      , 0               , XtRString, "black" },
+ {"bombreal", "XBomb", XtRPixel     , sizeof(String)      ,   sizeof(String), XtRString, "red" },
+ {"bombmark", "XBomb", XtRPixel     , sizeof(String)      , 2*sizeof(String), XtRString, "orange" },
+ {"number"  , "XBomb", XtRPixel     , sizeof(String)      , 3*sizeof(String), XtRString, "blue" },
+ {"hidden"  , "XBomb", XtRPixel     , sizeof(String)      , 4*sizeof(String), XtRString, "grey50" },
+ {"correct" , "XBomb", XtRPixel     , sizeof(String)      , 5*sizeof(String), XtRString, "green" },
+ {"font"    , "XBomb", XtRFontStruct, sizeof(XFontStruct*), 6*sizeof(String), XtRString, "lucidasans-24"}
 };
 
+
 /*++++++++++++++++++++++++++++++++++++++
   Open the X connection.
 
@@ -100,16 +113,17 @@ static XtResource xresources[]=
 
 void InitialiseX(int *argc,char **argv)
 {
+ Atom close_atom;
  XFontStruct *fontstruct;
  XGCValues values;
- Widget toplevel,form,quit,start,menu,menushell,hiscore;
+ Widget form,quit,start,menu,menushell;
  int i;
 
  /* Initialise the display */
 
  toplevel=XtVaAppInitialize(&app_context,"XBomb",
                             NULL, (Cardinal)0,argc,argv,NULL,
-                            XtNtitle,"XBomb V2    [(c) AMB Software 1994,95]",
+                            XtNtitle,"XBomb V2.1a",
                             XtNiconName,"XBomb",
                             NULL);
 
@@ -117,10 +131,14 @@ void InitialiseX(int *argc,char **argv)
 
  /* Create the widgets */
 
- form=XtVaCreateManagedWidget("form",formWidgetClass,toplevel,XtNwidth,100,XtNheight,100,NULL);
+ form=XtVaCreateManagedWidget("form",formWidgetClass,toplevel,
+                              XtNwidth,100,XtNheight,100,
+                              NULL);
+
+ XtAddEventHandler(form,KeyReleaseMask,False,(XtEventHandler)key_press_proc,NULL);
 
  start=XtVaCreateManagedWidget("start",commandWidgetClass,form,
-                               XtNlabel,"Start",
+                               XtNlabel,"Restart",
                                XtNfromHoriz,NULL,XtNfromVert,NULL,
                                XtNleft,XtChainLeft,XtNtop,XtChainTop,XtNright,XtChainLeft,XtNbottom,XtChainTop,
                                NULL);
@@ -160,7 +178,7 @@ void InitialiseX(int *argc,char **argv)
     XtAddCallback(item,XtNcallback,change_status_proc,(XtPointer)(GAME_TYPE+i));
    }
 
- hiscore=XtVaCreateManagedWidget("hiscore",commandWidgetClass,form,
+ hiscore=XtVaCreateManagedWidget("hiscore",toggleWidgetClass,form,
                                  XtNlabel,"Hi-Scores",
                                  XtNfromHoriz,menu,XtNfromVert,NULL,
                                  XtNleft,XtChainLeft,XtNtop,XtChainTop,XtNright,XtChainLeft,XtNbottom,XtChainTop,
@@ -182,7 +200,6 @@ void InitialiseX(int *argc,char **argv)
                                    NULL);
 
  XtAddEventHandler(play_area,ButtonPressMask,False,(XtEventHandler)mouse_press_proc,NULL);
- XtAddEventHandler(play_area,KeyReleaseMask,False,(XtEventHandler)key_press_proc,NULL);
  XtAddEventHandler(play_area,ExposureMask|StructureNotifyMask,False,(XtEventHandler)size_expose_proc,NULL);
 
  clock_w=XtVaCreateManagedWidget("clock",labelWidgetClass,form,
@@ -240,16 +257,59 @@ void InitialiseX(int *argc,char **argv)
  XtVaGetValues(play_area,XtNbackground,&values.foreground,NULL);
  gc[GC_BLANK]=XCreateGC(display,RootWindowOfScreen(XtScreen(toplevel)),GCForeground,&values);
 
+ /* The highscore table. */
+
+ highscore=XtVaCreatePopupShell("hiscores",topLevelShellWidgetClass,toplevel,
+                                XtNtitle,"XBomb V2.1a High Scores",XtNiconName,"XBomb Hi-scores",
+                                XtNmappedWhenManaged,False,
+                                NULL);
+
+ highscoresform=XtVaCreateManagedWidget("form",formWidgetClass,highscore,
+                                        XtNwidth,200,XtNheight,100,
+                                        NULL);
+
+ XtAddEventHandler(highscoresform,KeyReleaseMask,False,(XtEventHandler)key_press_proc,NULL);
+
+ for(i=0;i<12;i++)
+   {
+    char name[4][6];
+
+    sprintf(name[0],"hs%da",i);
+    sprintf(name[1],"hs%db",i);
+    sprintf(name[2],"hs%ds",i);
+    sprintf(name[3],"hs%dd",i);
+
+    highscores[i][0]=XtVaCreateManagedWidget(name[0],labelWidgetClass,highscoresform,
+                                             XtNborderWidth,0,XtNlabel,"MMM",
+                                             XtNfromHoriz,NULL,XtNfromVert,i?highscores[i-1][0]:NULL,
+                                             NULL);
+    highscores[i][1]=XtVaCreateManagedWidget(name[1],labelWidgetClass,highscoresform,
+                                             XtNborderWidth,0,XtNlabel,"MMMMMMM",
+                                             XtNfromHoriz,highscores[i][0],XtNfromVert,i?highscores[i-1][0]:NULL,
+                                             NULL);
+    highscores[i][2]=XtVaCreateManagedWidget(name[2],labelWidgetClass,highscoresform,
+                                             XtNborderWidth,0,XtNlabel,"MMMM",
+                                             XtNfromHoriz,highscores[i][1],XtNfromVert,i?highscores[i-1][0]:NULL,
+                                             NULL);
+    highscores[i][3]=XtVaCreateManagedWidget(name[3],labelWidgetClass,highscoresform,
+                                             XtNborderWidth,0,XtNlabel,"MMMMMMMMMMMMMM",
+                                             XtNjustify,XtJustifyLeft,
+                                             XtNfromHoriz,highscores[i][2],XtNfromVert,i?highscores[i-1][0]:NULL,
+                                             NULL);
+   }
+
  /* Add the icon */
 
  icon_p=XCreateBitmapFromData(display,RootWindowOfScreen(XtScreen(toplevel)),
                               icon_bits,icon_width,icon_height);
 
  XtVaSetValues(toplevel,XtNiconPixmap,icon_p,NULL);
+ XtVaSetValues(highscore,XtNiconPixmap,icon_p,NULL);
 
  /* Show the widgets */
 
  XtRealizeWidget(toplevel);
+ XtManageChild(highscore);
  XFlush(display);
 
  play_window=XtWindow(play_area);
@@ -260,6 +320,16 @@ void InitialiseX(int *argc,char **argv)
 
  xoffset=XTextWidth(fontstruct,"0",1)/2;
  yoffset=fontstruct->ascent/2;
+
+ /* Put an action on the close button */
+
+ close_atom=XInternAtom(display,"WM_DELETE_WINDOW",False);
+
+ XSetWMProtocols(display,XtWindow(toplevel),&close_atom,1);
+ XtAddEventHandler(toplevel,0,True,close_proc,NULL);
+
+ XSetWMProtocols(display,XtWindow(highscore),&close_atom,1);
+ XtAddEventHandler(highscore,0,True,close_proc,NULL);
 }
 
 
@@ -271,14 +341,20 @@ void FinishUpX(void)
 {
  int i;
 
+ /* Free the graphics contexts. */
+
  for(i=0;i<7;i++)
     XFreeGC(display,gc[i]);
 
+ /* Free the pixmaps. */
+
  for(i=0;i<2;i++)
     XFreePixmap(display,bombpix[i]);
 
  XFreePixmap(display,icon_p);
 
+ /* Close the display.. */
+
  XCloseDisplay(display);
 }
 
@@ -318,6 +394,8 @@ void DrawSquare(int x,int y,unsigned char value)
  static char grid_numbers[13][2]={" ","1","2","3","4","5","6","7","8","9","A","B","C"};
  int which_gc;
 
+ /* Find the position. */
+
  switch(grid_type)
    {
    case GAME_HEXAGON:
@@ -330,15 +408,17 @@ void DrawSquare(int x,int y,unsigned char value)
     ypos=y*scaley;
     break;
    case GAME_TRIANGLE:
-    xpos=x*scalex+scalex/2;
-    ypos=y*scaley*2-((x+y)%2)*scaley+scaley;
+    xpos=x*scalex/2+scalex/4;
+    ypos=y*scaley-((x+y)%2)*scaley/2+scaley/2;
     if((x+y)%2)
-       outline[4].x=xpos+scalex/2, outline[4].y=ypos+2*scaley;
+       outline[4].x=xpos+scalex/4, outline[4].y=ypos+scaley;
     else
-       outline[0].x=xpos+scalex/2, outline[0].y=ypos-scaley;
+       outline[0].x=xpos+scalex/4, outline[0].y=ypos-scaley/2;
     break;
    }
 
+ /* Choose the context. */
+
  if(value&UNSEEN)
    {
     if(value&THINK_BOMB)
@@ -354,6 +434,8 @@ void DrawSquare(int x,int y,unsigned char value)
        which_gc=GC_BLANK;
    }
 
+ /* Fill the background. */
+
  switch(grid_type)
    {
    case GAME_HEXAGON:
@@ -370,24 +452,24 @@ void DrawSquare(int x,int y,unsigned char value)
     break;
    }
 
+ /* Draw the bomb or text. */
+
  if(value&UNSEEN)
    {
     if(value&THINK_BOMB)
        XCopyArea(display,bombpix[1],play_window,gc[GC_THINK_BACK],0,0,pixw,pixh,xpos,ypos);
    }
- else
+ else if(value&ACTUAL_BOMB)
+    XCopyArea(display,bombpix[0],play_window,gc[GC_ACTUAL_BACK],0,0,pixw,pixh,xpos,ypos);
+ else if((value&~CORRECT)>0 && (value&~CORRECT)<=12)
    {
-    if(value&ACTUAL_BOMB)
-       XCopyArea(display,bombpix[0],play_window,gc[GC_ACTUAL_BACK],0,0,pixw,pixh,xpos,ypos);
-    else
-       if((value&~CORRECT)>0 && (value&~CORRECT)<=12)
-         {
-          int text_xpos=xpos+pixw/2-xoffset;
-          int text_ypos=ypos+pixh/2+yoffset;
-          XDrawString(display,play_window,gc[GC_NUMBERS],text_xpos,text_ypos,grid_numbers[(value&~CORRECT)],1);
-         }
+    int text_xpos=xpos+pixw/2-xoffset;
+    int text_ypos=ypos+pixh/2+yoffset;
+    XDrawString(display,play_window,gc[GC_NUMBERS],text_xpos,text_ypos,grid_numbers[(value&~CORRECT)],1);
    }
 
+ /* Draw the correct squares at the end. */
+
  if(value&CORRECT)
     which_gc=GC_CORRECT;
  else
@@ -412,6 +494,44 @@ void DrawSquare(int x,int y,unsigned char value)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Display the high scores.
+
+  char *scores[11][4] The score information.
+
+  int which_score Which score was just completed.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void DisplayHighScores(char *scores[11][4],int which_score)
+{
+ int i,j;
+ Pixel fore,back;
+
+ XtVaSetValues(highscores[0][0],XtNlabel,"Pos",NULL);
+ XtVaSetValues(highscores[0][1],XtNlabel,"Name",NULL);
+ XtVaSetValues(highscores[0][2],XtNlabel,"Time",NULL);
+ XtVaSetValues(highscores[0][3],XtNlabel,"Date",NULL);
+
+ XtVaGetValues(highscores[0][0],
+               XtNbackground,&back,XtNforeground,&fore,
+               NULL);
+
+ for(j=0;j<4;j++)
+    XtVaSetValues(highscores[0][j],
+                  XtNforeground,fore,
+                  XtNbackground,back,
+                  NULL);
+
+ for(i=0;i<11;i++)
+    for(j=0;j<4;j++)
+       XtVaSetValues(highscores[i+1][j],
+                     XtNforeground,i==which_score?back:fore,
+                     XtNbackground,i==which_score?fore:back,
+                     XtNlabel,scores[i][j],
+                     NULL);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   A callback that is activated by an expose event on the play area window.
 
@@ -442,88 +562,93 @@ static void size_expose_proc(Widget w,XtPointer va,XEvent* e,Boolean* vb)
 
 void ScaleWindow(void)
 {
- unsigned short width,height;
  unsigned short old_width,old_height;
  unsigned short new_width=0,new_height=0;
 
+ /* Get the old size. */
+
  XtVaGetValues(play_area,XtNwidth,&old_width,XtNheight,&old_height,NULL);
 
- new_width=old_width;
- new_height=old_height;
+ /* Choose a new size */
 
- do{
-    width=new_width;
-    height=new_height;
-
-    switch(grid_type)
-      {
-      case GAME_HEXAGON:
-       scalex=(int)width*2/(2*grid_width+1);
-       scaley=(int)height*4/(4*grid_height+1);
-       scalex=(int)(scalex<(scaley*8/7)?scalex:(scaley*8/7));
-       scaley=(scalex*7+8)/8;
-       new_width =scalex*(2*grid_width+1)/2;
-       new_height=scaley*(4*grid_height+1)/4;
-       break;
-      case GAME_SQUARE:
-       scalex=(int)width /grid_width;
-       scaley=(int)height/grid_height;
-       scalex=scaley=(int)(scalex<scaley?scalex:scaley);
-       new_width =scalex*grid_width;
-       new_height=scaley*grid_height;
-       break;
-      case GAME_TRIANGLE:
-       scalex=(int)width*2/(2*grid_width+1);
-       scaley=(int)height/(2*grid_height);
-       scalex=(int)(scalex<(scaley*8/7)?scalex:(scaley*8/7));
-       scaley=(scalex*7+8)/8;
-       new_width =scalex*grid_width+scalex;
-       new_height=2*scaley*grid_height;
-       break;
-      }
+ switch(grid_type)
+   {
+   case GAME_HEXAGON:
+    scalex=(int)old_width*2/(7*(2*grid_width+1)/6);
+    scaley=(int)old_height*3/(3*grid_height+1);
+    scaley=6*((scalex+scaley)/12);
+    if(scaley<30) scaley=18;
+    scalex=(scaley*7)/6;
+    new_width =scalex*(2*grid_width+1)/2;
+    new_height=scaley*(3*grid_height+1)/3;
+    break;
+   case GAME_SQUARE:
+    scalex=(int)old_width /grid_width;
+    scaley=(int)old_height/grid_height;
+    scalex=scaley=(scalex+scaley)/2;
+    if(scalex<30) scalex=scaley=20;
+    new_width =scalex*grid_width;
+    new_height=scaley*grid_height;
+    break;
+   case GAME_TRIANGLE:
+    scalex=(int)old_width*2/(grid_width+1);
+    scaley=(int)old_height/(7*grid_height/8);
+    scalex=4*((scalex+scaley)/8);
+    if(scalex<16) scalex=16;
+    scaley=(scalex*7)/8;
+    new_width =scalex/2*(grid_width+1);
+    new_height=scaley*grid_height;
+    break;
    }
- while(new_width!=width && new_height!=height);
+
+ /* If changed then re-size the window. */
 
  if(new_width!=old_width || new_height!=old_height)
    {
+    unsigned short width,height;
     new_width-=old_width;
     new_height-=old_height;
     XtVaGetValues(XtParent(XtParent(play_area)),XtNwidth,&width,XtNheight,&height,NULL);
     XResizeWindow(display,XtWindow(XtParent(XtParent(play_area))),(unsigned)(width+new_width),(unsigned)(height+new_height));
    }
- else
+ else /* Must be here due to a correct sizing. */
    {
     int i;
 
+    /* Get the outline of a funny shape tile. */
+
     if(grid_type==GAME_HEXAGON)
       {
        /* outline[0] is filled at the time of drawing */
-       outline[1].x= (int)scalex/2-1;   outline[1].y=-(int)scaley  /3;
-       outline[2].x= (int)scalex/2;     outline[2].y= (int)scaley  /3;
-       outline[3].x= 0;                 outline[3].y= (int)scaley*2/3;
-       outline[4].x=-(int)scalex/2;     outline[4].y= (int)scaley  /3;
-       outline[5].x=-(int)scalex/2+1;   outline[5].y=-(int)scaley  /3;
-       outline[6].x= 0;                 outline[6].y=-(int)scaley*2/3;
+       outline[1].x= (int)scalex/2;    outline[1].y=-(int)scaley/3;
+       outline[2].x= (int)scalex/2;    outline[2].y= (int)scaley/3;
+       outline[3].x= 0;                outline[3].y=-(int)scaley/3+scaley;
+       outline[4].x=-(int)scalex/2;    outline[4].y= (int)scaley/3;
+       outline[5].x=-(int)scalex/2;    outline[5].y=-(int)scaley/3;
+       outline[6].x= 0;                outline[6].y= (int)scaley/3-scaley;
       }
-    else
-       if(grid_type==GAME_TRIANGLE)
-         {
-          /* outline[0] is filled at the time of drawing */
-          outline[1].x=-(int)scalex+1;    outline[1].y= (int)scaley*2-1;
-          outline[2].x= (int)scalex*2-2;  outline[2].y=0;
-          outline[3].x=-(int)scalex+1;    outline[3].y=-(int)scaley*2+1;
-          /* outline[4] is filled at the time of drawing */
-          outline[5].x=-(int)scalex+1;    outline[5].y=-(int)scaley*2;
-          outline[6].x= (int)scalex*2-2;  outline[6].y=0;
-          outline[7].x=-(int)scalex+1;    outline[7].y= (int)scaley*2;
-         }
+    else if(grid_type==GAME_TRIANGLE)
+      {
+       /* outline[0] is filled at the time of drawing */
+       outline[1].x=-(int)scalex/2+1;    outline[1].y= (int)scaley-1;
+       outline[2].x= (int)scalex-2;      outline[2].y=0;
+       outline[3].x=-(int)scalex/2+1;    outline[3].y=-(int)scaley+1;
+       /* outline[4] is filled at the time of drawing */
+       outline[5].x=-(int)scalex/2+1;    outline[5].y=-(int)scaley;
+       outline[6].x= (int)scalex-2;      outline[6].y=0;
+       outline[7].x=-(int)scalex/2+1;    outline[7].y= (int)scaley;
+      }
+
+    /* Draw the bomb and flag pixmaps. */
 
     for(i=0;i<2;i++)
       {
-       if(grid_type==GAME_HEXAGON)
+       if(grid_type==GAME_SQUARE)
+          pixw=scalex,pixh=scaley;
+       else if(grid_type==GAME_HEXAGON)
           pixw=scalex,pixh=scaley*2/3;
        else
-          pixw=scalex,pixh=scaley;
+          pixw=scalex/2,pixh=scaley/2;
 
        if(bombpix[i])
           XFreePixmap(display,bombpix[i]);
@@ -551,11 +676,14 @@ void ScaleWindow(void)
          }
       }
 
+    /* Redraw the whole grid. */
+
     XClearWindow(display,play_window);
     DrawGrid();
    }
 }
 
+
 /*++++++++++++++++++++++++++++++++++++++
   A callback that is activated by a mouse press in the play area window.
 
@@ -576,6 +704,8 @@ static void mouse_press_proc(Widget w,XtPointer va,XEvent* e,Boolean* vb)
  int x_guess,y_guess,dx,dy;
  int i,j,d,dmin=1e6;
 
+ /* Work out which tile mouse pressed in. */
+
  switch(grid_type)
    {
    case GAME_HEXAGON:
@@ -598,14 +728,14 @@ static void mouse_press_proc(Widget w,XtPointer va,XEvent* e,Boolean* vb)
     y=e->xbutton.y/scaley;
     break;
    case GAME_TRIANGLE:
-    x_guess=e->xbutton.x/scalex;
-    y=e->xbutton.y/(2*scaley);
+    x_guess=e->xbutton.x*2/scalex;
+    y=e->xbutton.y/scaley;
     for(i=x_guess-1;i<=(x_guess+1);i++)
       {
        if(i<0 || i>=grid_width)
           continue;
-       dx=e->xbutton.x-(i*scalex+scalex);
-       dy=e->xbutton.y-(y*2*scaley+(1+(1+i+y)%2)*scaley*2/3);
+       dx=e->xbutton.x-(i*scalex/2+scalex/2);
+       dy=e->xbutton.y-(y*scaley+(1+(1+i+y)%2)*scaley/3);
        d=dx*dx+dy*dy;
        if(d<dmin)
          {x=i;dmin=d;}
@@ -613,9 +743,13 @@ static void mouse_press_proc(Widget w,XtPointer va,XEvent* e,Boolean* vb)
     break;
    }
 
+ /* Throw out errors. */
+
  if(x<0 || x>=grid_width || y<0 || y>=grid_height)
     {XBell(display,0);return;}
 
+ /* Do something based on button number. */
+
  switch(e->xbutton.button)
    {
    case 1: SelectSquare(x,y); break;
@@ -646,16 +780,18 @@ static void key_press_proc(Widget w,XtPointer va,XEvent* e,Boolean* vb)
 
  if(k==NoSymbol)return;
 
+ /* Do an action based on which key is pressed. */
+
  switch(k)
    {
    case XK_1:
-    status=GAME_LEVEL;
+    status=GAME_EASY;
     break;
    case XK_2:
-    status=GAME_LEVEL+1;
+    status=GAME_MEDIUM;
     break;
    case XK_3:
-    status=GAME_LEVEL+2;
+    status=GAME_DIFFICULT;
     break;
 
    case XK_s:
@@ -669,11 +805,18 @@ static void key_press_proc(Widget w,XtPointer va,XEvent* e,Boolean* vb)
     if(shift)
        status=GAME_HEXAGON;
     else
-       PrintHighScores(-1);
+       hiscore_proc(NULL,NULL,NULL);
     break;
 
    case XK_q:
-    status=GAME_QUIT;
+    do
+      {
+       if(w==toplevel)
+         {status=GAME_QUIT;break;}
+       if(w==highscore)
+         {hiscore_proc(NULL,NULL,NULL);break;}
+      }
+    while((w=XtParent(w)));
     break;
 
    case XK_t:
@@ -681,9 +824,8 @@ static void key_press_proc(Widget w,XtPointer va,XEvent* e,Boolean* vb)
        status=GAME_TRIANGLE;
     break;
 
-   default:
+   default: ;
    }
-
 }
 
 
@@ -704,6 +846,7 @@ static void change_status_proc(Widget widget,XtPointer clientData,XtPointer call
  status=(int)clientData;
 }
 
+
 /*++++++++++++++++++++++++++++++++++++++
   This function displays the hi-scores.
 
@@ -718,7 +861,51 @@ static void change_status_proc(Widget widget,XtPointer clientData,XtPointer call
 
 static void hiscore_proc(Widget widget,XtPointer clientData,XtPointer callData)
 {
- PrintHighScores(-1);
+ static Boolean state=False;
+
+ state=!state;
+
+ if(!widget)
+    XtVaSetValues(hiscore,XtNstate,state,NULL);
+
+ if(state)
+    XtPopup(highscore,XtGrabNone);
+ else
+    XtPopdown(highscore);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  A callback that is activated by a close window event on the toplevel window.
+
+  Widget w The widget that caused the callback.
+
+  XtPointer va Not used.
+
+  XEvent* e The event that requires action.
+
+  Boolean* vb Not used.
+
+  This function is only ever called from the Xt Intrinsics routines.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void close_proc(Widget w,XtPointer va,XEvent* e,Boolean* vb)
+{
+ XClientMessageEvent *cev=(XClientMessageEvent*)e;
+ Atom atom_type=XInternAtom(display,"WM_PROTOCOLS",False);
+
+ if(atom_type==cev->message_type)
+   {
+    Atom atom_proto=XInternAtom(display,"WM_DELETE_WINDOW",False);
+
+    if(cev->format==32 && atom_proto==(Atom)cev->data.l[0])
+      {
+       if(w==toplevel)
+          status=GAME_QUIT;
+       else
+          hiscore_proc(NULL,NULL,NULL);
+      }
+   }
 }
 
 
